/// This module contains everything everything, to convert the given JSON file into Rust structs using serde.
mod parser;

use log::info;
use serde::Deserialize;
use time::{format_description, OffsetDateTime};

use crate::{overview::parser::OgdSearchResult, Error};

/// Returns the current date in YYYY-MM-DD format. Needed for RIS API query to get current version of the overview.
fn current_date() -> String {
    let local_date = OffsetDateTime::now_utc();
    let format = format_description::parse("[year]-[month]-[day]").unwrap(); //unwrap okay, supplied format is fine
    local_date.format(&format).expect("Failed to format date")
}

/// Fetches the json content of the given overview (`law_id`) from the RIS API.
///
/// # Errors
/// Fails if `ureq` can't create a connection, probably because there's no internet connection? (Or RIS is not online.)
fn fetch_page(overview_id: usize, page: usize) -> Result<String, Error> {
    Ok(
        ureq::post("https://data.bka.gv.at/ris/api/v2.6/Bundesrecht")
            .send_form(&[
                ("Applikation", "BrKons"),
                ("Gesetzesnummer", &format!("{}", overview_id)),
                ("DokumenteProSeite", "OneHundred"),
                ("Seitennummer", &format!("{}", page)),
                ("Fassung.FassungVom", &current_date()),
            ])?
            .into_string()?,
    )
}

#[derive(Deserialize)]
#[serde(rename_all = "PascalCase")]
pub(crate) struct Wrapper {
    ogd_search_result: OgdSearchResult,
}

pub(crate) fn parse(overview_id: usize) -> Result<Vec<String>, Error> {
    let mut page = 1;
    let mut skip = true;
    let mut ret = Vec::new();
    loop {
        info!("=== Fetching overview page #{page} ===");
        let json = fetch_page(overview_id, page)?;
        let (cont, nodes) = parse_from_str(&json, skip)?;
        for n in nodes {
            ret.push(n.clone());
        }
        if !cont {
            break;
        }
        skip = false;
        page += 1;
    }

    Ok(ret)
}

pub(crate) fn parse_from_str(
    content: &str,
    skip_first: bool,
) -> Result<(bool, Vec<String>), Error> {
    let mut ret = Vec::new();
    let wrapper: Wrapper = serde_json::from_str(content)?;

    let iter = wrapper.ogd_search_result.get_par().into_iter();
    let boxed_iter: Box<dyn Iterator<Item = String>> = if skip_first {
        Box::new(iter.skip(1))
    } else {
        Box::new(iter)
    };
    for par in boxed_iter {
        ret.push(par);
        //if !crate::par::parse(&par).unwrap() {
        //    return Ok(false);
        //}
    }

    if !wrapper.ogd_search_result.has_next_page() {
        return Ok((false, ret));
    }
    Ok((true, ret))
}
